package FeldmanVerifiableSecretSharing;

import FeldmanVerifiableSecretSharing.ShamirSecretSharing.Polynomial;
import FeldmanVerifiableSecretSharing.ShamirSecretSharing.SecretSharing;
import org.bouncycastle.util.Arrays;
import org.factcenter.qilin.primitives.CyclicGroup;
import org.factcenter.qilin.primitives.concrete.Zpstar;

import java.math.BigInteger;
import java.util.Random;

/**
 * Created by Tzlil on 1/27/2016.
 *
 * an implementation of Feldman's verifiable secret sharing scheme.
 *
 * allows trusted dealer to share a key x among n parties.
 *
 */
public class VerifiableSecretSharing extends SecretSharing {
    private final Zpstar zpstar;
    private final BigInteger g; // public generator of group
    private final BigInteger[] commitments;
    private final BigInteger y; // y = g ^ x

    /**
     * @param p a large prime
     * @param q a large prime dividing p - 1.
     * @param g a generator of cyclic group of order q.
     *          the generated group is a subgroup of Zp*.
     *          it must be chosen such that computing discrete logarithms is hard in this group.
     */
    public VerifiableSecretSharing(int t, int n, BigInteger x, Random random,BigInteger p,BigInteger q,BigInteger g) {
        super(t, n, x, random,q);
        this.g = g;
        this.zpstar = new Zpstar(p);
        assert (zpstar.contains(g));
        assert (p.subtract(BigInteger.ONE).mod(q).equals(BigInteger.ZERO)); // assert p - 1 % q == 0
        this.commitments = generateCommitments();
        this.y = zpstar.multiply(g,x);
    }

    /**
     * @return commitments[i] = g ^ polynomial.coefficients[i]
     */
    private BigInteger[] generateCommitments() {

        Polynomial polynomial = getPolynomial();
        BigInteger[] coefficients = polynomial.getCoefficients();
        BigInteger[] commitments = new BigInteger[coefficients.length];
        for (int i = 0 ; i < commitments.length;i++){
            commitments[i] = zpstar.multiply(g,coefficients[i]);
        }
        return commitments;
    }

    /**
     * @param i share holder id
     * @param commitments
     * @param zpstar
     *
     * @return product of commitments[j] ^ (i ^ j) == g ^ polynomial(i)
     */
    public static BigInteger verify(int i,BigInteger[] commitments,Zpstar zpstar) {
        BigInteger v = zpstar.zero();
        BigInteger power = BigInteger.ONE;
        BigInteger I = BigInteger.valueOf(i);
        for (int j = 0 ; j < commitments.length ; j ++){
            v = zpstar.add(v,zpstar.multiply(commitments[j],power));
            power = power.multiply(I);
        }
        return v;
    }


    /**
     * getter
     * @return generator of group
     */
    public BigInteger getGenerator() {
        return g;
    }

    /**
     * getter
     * @return copy of commitments
     */
    public BigInteger[] getCommitments() {
        return Arrays.clone(commitments);
    }

    /**
     * getter
     * @return zpstar
     */
    public Zpstar getZpstar(){
        return zpstar;
    }

    /**
     * getter
     * @return public value of this
     */
    public BigInteger getY(){
        return y;
    }
}
